import os
import re
from timeit import default_timer
from typing import Tuple

from fastapi import FastAPI
from prometheus_client import Histogram
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Match


class PrometheusFastApiInstrumentator:
    def __init__(
        self,
        should_group_status_codes: bool = True,
        should_ignore_untemplated: bool = False,
        should_group_untemplated: bool = True,
        should_round_latency_decimals: bool = False,
        should_respect_env_var_existence: bool = False,
        excluded_handlers: list = ["/metrics"],
        buckets: tuple = Histogram.DEFAULT_BUCKETS,
        metric_name: str = "http_request_duration_seconds",
        label_names: tuple = ("method", "handler", "status",),
        round_latency_decimals: int = 4,
        env_var_name: str = "PROMETHEUS",
    ):
        """
        :param should_group_status_codes: Should status codes be grouped into 
            `2xx`, `3xx` and so on?

        :param should_ignore_untemplated: Should requests without a matching 
            template be ignored?

        :param should_group_untemplated: Should requests without a matching 
            template be grouped to handler `none`?

        :param should_round_latency_decimals: Should recorded latencies be 
            rounded to a certain number of decimals?

        :param should_respect_env_var_existence: Should the instrumentator only 
            work - for example the methods `instrument()` and `expose()` - if 
            a certain environment variable is set? Usecase: A base FastAPI app 
            that is used by multiple distinct apps. The apps only have to set 
            the variable to be instrumented.

        :param excluded_handlers: Handlers that should be ignored. List of 
            strings is turned into regex patterns.

        :param buckets: Buckets for the histogram. Defaults to Prometheus default.

        :param metric_name: Name of the latency metric.

        :param label_names: Names of the three labels used for the metric. Does 
            not influence the label values. Defaults to ("method", "handler", 
            "status",).

        :param round_latency_decimals: Number of decimals latencies should be 
            rounded to. Ignored unless `should_round_latency_decimals` is `True`.

        :param env_var_name: Any valid os environment variable name that 
            will be checked for existence before instrumentation. Ignored 
            unless `should_respect_env_var_existence` is `True`.
        """

        self.should_group_status_codes = should_group_status_codes
        self.should_ignore_untemplated = should_ignore_untemplated
        self.should_group_untemplated = should_group_untemplated
        self.should_round_latency_decimals = should_round_latency_decimals
        self.should_respect_env_var_existence = should_respect_env_var_existence

        self.round_latency_decimals = round_latency_decimals
        self.label_names = label_names
        self.env_var_name = env_var_name

        if excluded_handlers:
            self.excluded_handlers = [re.compile(path) for path in excluded_handlers]
        else:
            self.excluded_handlers = []

        if buckets[-1] == float("inf"):
            self.buckets = buckets
        else:
            self.buckets = buckets + (float("inf"),)

        self.histogram = Histogram(
            name=metric_name,
            documentation="Duration of HTTP requests in seconds",
            labelnames=self.label_names,
            buckets=self.buckets,
        )

    def instrument(self, app: FastAPI) -> "self":
        """Performs the instrumentation by adding middleware and endpoint.
        
        :param app: FastAPI app to be instrumented.
        :param return: self.
        """

        if self.should_respect_env_var_existence and self.env_var_name not in os.environ:
            return self

        @app.middleware("http")
        async def dispatch_middleware(request: Request, call_next) -> Response:
            start_time = default_timer()

            method = request.method

            handler, is_templated = self._get_handler(request)

            try:
                response = None
                response = await call_next(request)
            except Exception as e:
                if response is None:
                    status = 500
                raise e from None
            else:
                status = response.status_code

            if self._is_handler_excluded(handler, is_templated):
                return response

            if is_templated is False and self.should_group_untemplated:
                handler = "none"

            duration = max(default_timer() - start_time, 0)

            if self.should_round_latency_decimals:
                duration = round(duration, self.round_latency_decimals)

            self.histogram.labels(
                *self._create_label_tuple(method, handler, status)
            ).observe(duration)

            return response

        return self

    def expose(self, app: FastAPI, endpoint: str = "/metrics") -> "self":
        """Exposes Prometheus metrics by adding endpoint to the given app.

        **Important**: There are many different ways to expose metrics. This is 
        just one of them, suited for both multiprocess and singleprocess mode. 
        Refer to the Prometheus Python client documentation for more information.

        :param app: FastAPI where the endpoint should be added to.
        :param endpoint: Route of the endpoint. Defaults to "/metrics".
        :param return: self.
        """

        if self.should_respect_env_var_existence and self.env_var_name not in os.environ:
            return self

        from prometheus_client import (CONTENT_TYPE_LATEST, REGISTRY,
                                       CollectorRegistry, generate_latest,
                                       multiprocess)

        if "prometheus_multiproc_dir" in os.environ:
            pmd = os.environ["prometheus_multiproc_dir"]
            if os.path.isdir(pmd):
                registry = CollectorRegistry()
                multiprocess.MultiProcessCollector(registry)
            else:
                raise ValueError(
                    f"Env var prometheus_multiproc_dir='{pmd}' not a directory."
                )
        else:
            registry = REGISTRY

        @app.get("/metrics")
        def metrics():
            return Response(generate_latest(registry), media_type=CONTENT_TYPE_LATEST)

        return self

    def _create_label_tuple(
        self, method: str, handler: str, code: int
    ) -> Tuple[str, str, str]:
        """Processes label values based on config.

        :param method: Method used for the request, for example `GET`.
        :param handler: Identifier for entity that handled / should handle the 
            request.
        :param code: Status code of the response. If error / exception occured 
            this should be `500`.
        :return: Processed values.
        """

        code = str(code)

        if self.should_group_status_codes:
            code = code[0] + "xx"
        return (
            method,
            handler,
            code,
        )

    def _get_handler(self, request: Request) -> Tuple[str, bool]:
        """Extracts either template or (if no template) path."""

        for route in request.app.routes:
            match, child_scope = route.matches(request.scope)
            if match == Match.FULL:
                return route.path, True

        return request.url.path, False

    def _is_handler_excluded(self, handler: str, is_templated: bool) -> bool:
        """Determines if the handler should be ignored."""

        if is_templated is False and self.should_ignore_untemplated:
            return True

        if any(pattern.search(handler) for pattern in self.excluded_handlers):
            return True

        return False



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['prometheus_fastapi_instrumentator']

package_data = \
{'': ['*']}

install_requires = \
['fastapi>=0.38.1,<=1.0.0', 'prometheus-client>=0.8.0,<0.9.0']

setup_kwargs = {
    'name': 'prometheus-fastapi-instrumentator',
    'version': '1.3.0',
    'description': 'Instrument your FastAPI with Prometheus metrics',
    'long_description': '# Prometheus FastAPI Instrumentator\n\n[![PyPI version](https://badge.fury.io/py/prometheus-fastapi-instrumentator.svg)](https://pypi.python.org/pypi/prometheus-fastapi-instrumentator/)\n[![Maintenance](https://img.shields.io/badge/maintained%3F-yes-green.svg)](https://GitHub.com/Naereen/StrapDown.js/graphs/commit-activity)\n[![downloads](https://img.shields.io/pypi/dm/prometheus-fastapi-instrumentator)](https://pypi.org/project/prometheus-fastapi-instrumentator/)\n\n![release](https://github.com/trallnag/prometheus-fastapi-instrumentator/workflows/release/badge.svg)\n![test branches](https://github.com/trallnag/prometheus-fastapi-instrumentator/workflows/test%20branches/badge.svg)\n[![codecov](https://codecov.io/gh/trallnag/prometheus-fastapi-instrumentator/branch/master/graph/badge.svg)](https://codecov.io/gh/trallnag/prometheus-fastapi-instrumentator)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\nInstrument your FastAPI with Prometheus metrics. Install with:\n\n    pip install prometheus-fastapi-instrumentator\n\n## Fast Track\n\n```python\nfrom prometheus_fastapi_instrumentator import Instrumentator\nInstrumentator().instrument(app).expose(app)\n```\n\nWith this single line FastAPI is instrumented and all Prometheus metrics used \nin the FastAPI app can be scraped via the added `/metrics` endpoint. \n\nThe exporter includes the single metric `http_request_duration_seconds` of \nthe type Histogram. A separate `http_requests_total` isn\'t necessary as the \ntotal can be retrieved with the `http_requests_total_count` series.\n\nThe Prometheus FastAPI Instrumentator (any idea for a short hand?) is highly\nconfigurable and has few handy features.\n\n* **Opt-out** (activated by default):\n    * Status codes are grouped into `2xx`, `3xx` and so on.\n    * Requests without a matching template are grouped into the handler `none`.\n    * Regex patterns to ignore certain routes.    \n* **Opt-in** (Deactivated by default):\n    * Control instrumentation and exposition of FastAPI at runtime by setting \n        an environment variable.\n    * Rounding of latencies to a certain decimal number.\n    * Completely ignore untemplated routes.\n    * Renaming of labels and the metric.\n\n\nSee the *Example with all parameters* for all possible options.\n\n## Example with all parameters\n\n```python\nfrom prometheus_fastapi_instrumentator import PrometheusFastApiInstrumentator\nPrometheusFastApiInstrumentator(\n    should_group_status_codes=False,\n    should_ignore_untemplated=True,\n    should_group_untemplated=False,\n    should_round_latency_decimals=True,\n    should_respect_env_var_existence=True,\n    excluded_handlers=["/metrics", "/admin"],\n    buckets=[1, 2, 3, 4, 5],\n    metric_name="my_custom_metric_name",\n    label_names=("method_type", "path", "status_code",),\n    round_latency_decimals=3,\n    env_var_name="PROMETHEUS",\n).instrument(app).expose(app, "/prometheus_metrics")\n```\n\n`instrument`: Instruments the given FastAPI based on the configuration in \nthe constructur of the exporter class.\n\n`expose`: Completely separate from `instrument` and not necessary for \ninstrumentation. Just a simple option to expose metrics by adding an endpoint \nto the given FastAPI. Supports multiprocess mode.\n\n## Prerequesites\n\n* `python = "^3.6"` (tested with 3.6 and 3.8)\n* `fastapi = ">=0.38.1, <=1.0.0"` (tested with 0.38.1 and 0.59.0)\n* `prometheus-client = "^0.8.0"` (tested with 0.8.0)\n\n## Development\n\nDeveloping and building this package on a local machine requires \n[Python Poetry](https://python-poetry.org/). I recommend to run Poetry in \ntandem with [Pyenv](https://github.com/pyenv/pyenv). Once the repository is \ncloned, run `poetry install` and `poetry shell`. From here you may start the \nIDE of your choice.\n\nFor formatting, the [black formatter](https://github.com/psf/black) is used.\nRun `black .` in the repository to reformat source files. It will respect\nthe black configuration in the `pyproject.toml`. For more information just \ntake a look at the GitHub workflow files.\n\n',
    'author': 'Tim Schwenke',
    'author_email': 'tim.schwenke+github@protonmail.ch',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/trallnag/prometheus-fastapi-instrumentator',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
